
/* Organize ROM as pairwise 32-bit wide memories to allow access to 48-bit wrapped values 
   |------------16------------|
   |_____|_____|  |_____|_____|
   |----15-----|  |-13--|-13--|

*/

module ROMArbiter (clk,
				   reset,
				
				   ch0vi,
				   ch0so,
				   ch0addr,
				   ch0data,
				   ch0vo,
				   ch0si,
				   ch0tagin,
				   ch0tagout,
				   
				   ch1vi,
				   ch1so,
				   ch1addr,
				   ch1data,
				   ch1tagin,
				   ch1tagout,
				   ch1vo,
				   ch1si,
				   
				   addrlo,
				   addrhi,
				   datalo,
				   datahi);
	
	parameter ADDRESS_WIDTH = 16;
	parameter DATA_WIDTH = 48;
	parameter ROM_WIDTH = 32;
	parameter ROM_ADDRESS_WIDTH = 13;
	
	input clk;
	input reset;
	
	input ch0vi;		
	output ch0so;						//reg ch0so;
	input [ADDRESS_WIDTH-1:0] ch0addr;
	output [DATA_WIDTH-1:0] ch0data;
	output ch0vo;						reg ch0vo;
	input ch0si;
	input ch0tagin;
	output ch0tagout;					reg ch0tagout;
		
	input ch1vi;
	output ch1so;						//reg ch1so;
	input [ADDRESS_WIDTH-1:0] ch1addr;
	output [DATA_WIDTH-1:0] ch1data;
	output ch1vo;						reg ch1vo;
	input ch1si;
	input ch1tagin;
	output ch1tagout;					reg ch1tagout;
		
	output [ROM_ADDRESS_WIDTH-1:0] addrlo;
	output [ROM_ADDRESS_WIDTH-1:0] addrhi;
	input [ROM_WIDTH-1:0] datalo;
	input [ROM_WIDTH-1:0] datahi;
	

	// channel 0 has priority
	// no request on ch1 or ch0 cannot proceed because full and stalled
	wire ch1proceed = ch1vi & (~ch0vi | (ch0vo & ch0si));
	wire [ADDRESS_WIDTH-1:0] addr = ch1proceed ? ch1addr : ch0addr;
	
	assign addrlo = addr[2] ? addr[ADDRESS_WIDTH-1:3] + 1 : addr[ADDRESS_WIDTH-1:3];
	assign addrhi = addr[ADDRESS_WIDTH-1:3];
	
	// 64 bits
	reg [(2*ROM_WIDTH)-1:0] data;
	reg [(2*ROM_WIDTH)-1:0] data0;
	reg [(2*ROM_WIDTH)-1:0] data1;
	
	assign ch0data = data0[DATA_WIDTH-1:0];
	assign ch1data = data1[DATA_WIDTH-1:0];
	
	assign ch0so = (ch0vi & ch0vo & ch0si);
	assign ch1so = (ch0vi & ch1vi & (~ch0vo | ~ch0si)) | (ch1vi & ch1vo & ch1si);
	
`ifdef SIM
	reg ch0vifu;
	reg ch1vifu;

	always @ (clk or ch0vi or ch1vi) begin
		if (!clk) begin
			ch0vifu = ch0vi;
			ch1vifu = ch1vi;
		end
	end
`endif


	always @ (posedge clk) begin
		if (!reset) begin
			ch0vo = 1'b0;
//			ch0so = 1'b0;
			
			ch1vo = 1'b0;
//			ch1so = 1'b0;
		end
		else begin			
			data = addr[2] ? {datalo, datahi} : {datahi, datalo};
	
			case (addr[1:0])
				2'd0: ;
					 
				2'd1: // align up one byte
						data = {8'd0, data[DATA_WIDTH-1+8:8]};
					  
				2'd2: // align up one word
						data = {16'd0, data[DATA_WIDTH-1+16:16]};
					  
				2'd3: // align up 3 bytes and pad
						data = {24'd0, data[(2*ROM_WIDTH)-1:24]};				  
			endcase


			// clear valid out signals if not stalled
			if (!ch0si)
				ch0vo = 1'b0;
			
			if (!ch1si)
				ch1vo = 1'b0;


//			ch0so = 0;
			
`ifdef SIM
			if (ch0vifu) begin
`else
			if (ch0vi) begin
`endif
				// stall if we have output and stall input and request incoming
//				ch0so = ch0si & ch0vo;
				
				if (!ch0si | !ch0vo) begin
					ch0vo = 1;
					ch0tagout = ch0tagin;
					data0 = data;
				end

			end			


//			ch1so = 0;
			
`ifdef SIM
			if (ch1vifu) begin
`else
			if (ch1vi) begin
`endif
				// stall if we have output and stall input and request incoming
				// or if channel 0 is going to proceed
				if (ch1si & ch1vo) begin
//					ch1so = 1'b1;
				end
				else if (ch0vi & (!ch0vo | !ch0si)) begin
//					ch1so = 1'b1;
				end
				// otherwise we can proceed
				else begin
					ch1vo = 1;
					ch1tagout = ch1tagin;
					data1 = data;
				end
			end
		end	
	end	
endmodule		
	


module ROM (en,
			addr,
			dataOut);
			
	parameter ADDRESS_WIDTH = 16;
	parameter DATA_WIDTH = 16;
	parameter DEPTH = 16;
	parameter FILE = "rom.data";
	
	input en;
	
	input [ADDRESS_WIDTH-1:0] addr;
	
	output [DATA_WIDTH-1:0] dataOut;
	
	reg [DATA_WIDTH-1:0] data [DEPTH-1:0];
	
	assign dataOut = data[addr];
		
	initial begin
		$readmemh(FILE, data);
	end
endmodule
	
	
