
`define SIM


`define INACTIVE	3'b000
`define START		3'b010
`define RUNNING		3'b011
`define IDLE		3'b100
`define INTERRUPT	3'd6
`define RESET		3'd7


`define THREAD0_PC 		16'h0110
`define THREAD1_PC 		16'h0111
`define THREAD2_PC 		16'h0112
`define THREAD3_PC 		16'h0113

`define THREAD0_SP 		16'h0114
`define THREAD1_SP 		16'h0115
`define THREAD2_SP 		16'h0116
`define THREAD3_SP 		16'h0117

`define THREAD0_ACTIVE 	16'h0118
`define THREAD1_ACTIVE 	16'h0119
`define THREAD2_ACTIVE 	16'h011A
`define THREAD3_ACTIVE 	16'h011B


module Threader2(clk,
				 reset,
				 
				 en,
				 we,
				 address,
				 datain,
				 
				 thread,
				 pc,
				 sp,
				 sr,
				 run_vo,
				 run_si,
				 
				 vector,
				 int_vo,
				 int_si,
				 
				 done,
				 pcn,
				 spn,
				 srn,
				 threadn);
		
	parameter ADDRESS_WIDTH = 16;
	parameter DATA_WIDTH = 16;
	parameter THREAD_COUNT = 2;
	

	input clk;
	input reset;
	
	input en;
	input we;
	input [ADDRESS_WIDTH-1:0] address;
	input [DATA_WIDTH-1:0] datain;
	
	output thread;						reg thread;
	output [ADDRESS_WIDTH-1:0] pc;		reg [ADDRESS_WIDTH-1:0] pc;
	output [ADDRESS_WIDTH-1:0] sp;		reg [ADDRESS_WIDTH-1:0] sp;
	output [DATA_WIDTH-1:0] sr;			reg [ADDRESS_WIDTH-1:0] sr;

	output run_vo;						reg run_vo;
	input run_si;
	
	output [3:0] vector;				reg [3:0] vector;
	output int_vo;						reg int_vo;
	input int_si;
	
	input done;
	input [ADDRESS_WIDTH-1:0] pcn;
	input [ADDRESS_WIDTH-1:0] spn;
	input [DATA_WIDTH-1:0] srn;
	input threadn;
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// State
	reg [2:0] state [THREAD_COUNT-1:0];
	
	wire [2:0] state_thread0 = state[0];
	wire [2:0] state_thread1 = state[1];
	wire [2:0] state_thread2 = state[2];
	wire [2:0] state_thread3 = state[3];
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// Contexts
	reg [ADDRESS_WIDTH-1:0] pcs [THREAD_COUNT-1:0];
	reg [ADDRESS_WIDTH-1:0] sps [THREAD_COUNT-1:0];
	reg [ADDRESS_WIDTH-1:0] srs [THREAD_COUNT-1:0];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread0 = pcs[0];
	wire [ADDRESS_WIDTH-1:0]  sps_thread0 = sps[0];
	wire [ADDRESS_WIDTH-1:0]  srs_thread0 = srs[0];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread1 = pcs[1];
	wire [ADDRESS_WIDTH-1:0]  sps_thread1 = sps[1];
	wire [ADDRESS_WIDTH-1:0]  srs_thread1 = srs[1];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread2 = pcs[2];
	wire [ADDRESS_WIDTH-1:0]  sps_thread2 = sps[2];
	wire [ADDRESS_WIDTH-1:0]  srs_thread2 = srs[2];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread3 = pcs[3];
	wire [ADDRESS_WIDTH-1:0]  sps_thread3 = sps[3];
	wire [ADDRESS_WIDTH-1:0]  srs_thread3 = srs[3];


	//////////////////////////////////////////////////////////////////////////////////////////////
	// Thread Selection (2-threads)
	reg next_thread;
	//	
	//////////////////////////////////////////////////////////////////////////////////////////////

	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// Prop Tracking
`ifdef SIM
	reg [32:0] icount;
`endif	
	//	
	//////////////////////////////////////////////////////////////////////////////////////////////

	integer i;
	always @ (posedge clk) begin
		if (!reset) begin
`ifdef SIM
			icount = 32'd0;
`endif	
			
			thread = 0;
			pc = 0;
			sp = 0;
			sr = 0;
			
			for (i = 0; i < THREAD_COUNT; i = i + 1) begin
				state[i] = `INACTIVE;
				pcs[i] = 0;
				srs[i] = 0;
				sps[i] = 0;
			end

			run_vo = 1'b0;

			// Reset triggers interrupt in first thread
			state[0] = `INTERRUPT;
			int_vo = 1'b1;

			pc = 16'hFFFE;
			vector = 4'd15;
		end
		else begin
			// Update from completed instructions first
			if (done) begin
				pcs[threadn] = pcn;
				sps[threadn] = spn;
				srs[threadn] = srn;
				// Leave inactive threads inactive
				if (state[threadn] != `INACTIVE)
					state[threadn] = `IDLE;			

`ifdef SIM
				icount = icount + 1;
`endif	
			end

			if (!run_si)
				run_vo = 1'b0;
			int_vo = 1'b0;

			if (state[thread] == `INTERRUPT) begin
				int_vo = 1'b1;
				state[thread] = `RUNNING;
			end 
			else begin
				// Launch thread if target is idle
				if (!run_si) begin
					next_thread = ~thread;

					if (state[next_thread] == `IDLE)
						thread = next_thread;

					if (state[thread] == `IDLE) begin
						pc = pcs[thread];
						sp = sps[thread];
						sr = srs[thread];
						
						run_vo = 1'b1;
						
						state[thread] = `RUNNING;
					end
				end
			end

			// Handle memory mapped interface
			if (en && we) begin
				if (address == `THREAD0_PC) begin
	`ifdef SIM
					$display("setting pc0 = %h", datain);
	`endif
					pcs[0] = datain;
				end
				else if (address == `THREAD1_PC) begin
	`ifdef SIM
					$display("setting pc1 = %h", datain);
	`endif
					pcs[1] = datain;
				end
				else if (address == `THREAD0_SP) begin
	`ifdef SIM
					$display("setting sp0 = %h", datain);
	`endif
					sps[0] = datain;
				end
				else if (address == `THREAD1_SP) begin
	`ifdef SIM
					$display("setting sp1 = %h", datain);
	`endif
					sps[1] = datain;
				end
				else if (address == `THREAD0_ACTIVE) begin
					if (datain == 16'd0) begin
						state[0] = `INACTIVE;
					end
					else if (state[0] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread0");
	`endif
						state[0] = `IDLE;
					end
				end
				else if (address == `THREAD1_ACTIVE) begin
					if (datain == 16'd0) begin
						state[1] = `INACTIVE;
					end
					else if (state[1] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread1");
	`endif
						state[1] = `IDLE;
					end
				end
			end
		end
	end
	//
	//////////////////////////////////////////////////////////////////////////////////////////////
	
endmodule




module Threader4(clk,
				 reset,
				 
				 en,
				 we,
				 address,
				 datain,
				 
				 thread,
				 pc,
				 sp,
				 sr,
				 run_vo,
				 run_si,
				 
				 vector,
				 int_vo,
				 int_si,
				 
				 done,
				 pcn,
				 spn,
				 srn,
				 threadn);
		
	parameter ADDRESS_WIDTH = 16;
	parameter DATA_WIDTH = 16;
	parameter THREAD_COUNT = 4;
	

	input clk;
	input reset;
	
	input en;
	input we;
	input [ADDRESS_WIDTH-1:0] address;
	input [DATA_WIDTH-1:0] datain;
	
	output [1:0] thread;				reg [1:0] thread;
	output [ADDRESS_WIDTH-1:0] pc;		reg [ADDRESS_WIDTH-1:0] pc;
	output [ADDRESS_WIDTH-1:0] sp;		reg [ADDRESS_WIDTH-1:0] sp;
	output [DATA_WIDTH-1:0] sr;			reg [ADDRESS_WIDTH-1:0] sr;

	output run_vo;						reg run_vo;
	input run_si;
	
	output [3:0] vector;				reg [3:0] vector;
	output int_vo;						reg int_vo;
	input int_si;
	
	input done;
	input [ADDRESS_WIDTH-1:0] pcn;
	input [ADDRESS_WIDTH-1:0] spn;
	input [DATA_WIDTH-1:0] srn;
	input [1:0] threadn;
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// State
	reg [2:0] state [THREAD_COUNT-1:0];
	
	wire [2:0] state_thread0 = state[0];
	wire [2:0] state_thread1 = state[1];
	wire [2:0] state_thread2 = state[2];
	wire [2:0] state_thread3 = state[3];
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// Contexts
	reg [ADDRESS_WIDTH-1:0] pcs [THREAD_COUNT-1:0];
	reg [ADDRESS_WIDTH-1:0] sps [THREAD_COUNT-1:0];
	reg [ADDRESS_WIDTH-1:0] srs [THREAD_COUNT-1:0];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread0 = pcs[0];
	wire [ADDRESS_WIDTH-1:0]  sps_thread0 = sps[0];
	wire [ADDRESS_WIDTH-1:0]  srs_thread0 = srs[0];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread1 = pcs[1];
	wire [ADDRESS_WIDTH-1:0]  sps_thread1 = sps[1];
	wire [ADDRESS_WIDTH-1:0]  srs_thread1 = srs[1];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread2 = pcs[2];
	wire [ADDRESS_WIDTH-1:0]  sps_thread2 = sps[2];
	wire [ADDRESS_WIDTH-1:0]  srs_thread2 = srs[2];
	
	wire [ADDRESS_WIDTH-1:0]  pcs_thread3 = pcs[3];
	wire [ADDRESS_WIDTH-1:0]  sps_thread3 = sps[3];
	wire [ADDRESS_WIDTH-1:0]  srs_thread3 = srs[3];

	//////////////////////////////////////////////////////////////////////////////////////////////
	// Thread Selection (2-threads)
	reg [1:0] next_thread;
	//	
	//////////////////////////////////////////////////////////////////////////////////////////////

	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// Prop Tracking
`ifdef SIM
	reg [31:0] icount;
	reg [31:0] icount_thread0;
	reg [31:0] icount_thread1;
	reg [31:0] icount_thread2;
	reg [31:0] icount_thread3;
	reg [31:0] num_clock;
	
	reg [31:0] num_clock_thread0;
	reg [31:0] num_clock_thread1;
	reg [31:0] num_clock_thread2;
	reg [31:0] num_clock_thread3;
	
	reg [5:0] count_per_instr;
	
	reg [31:0] latency_thread0;
	reg [31:0] latency_thread1;
	reg [31:0] latency_thread2;
	reg [31:0] latency_thread3;
	
	reg [31:0] latency_total;
	
	
	
`endif	

    
	//	
	//////////////////////////////////////////////////////////////////////////////////////////////

	integer i;
	always @ (posedge clk) begin
		if (!reset) begin
`ifdef SIM
			icount = 32'd0;
			num_clock = 32'd0;
            icount_thread0 = 32'd0;
            icount_thread1 = 32'd0;
            icount_thread2 = 32'd0;
            icount_thread3 = 32'd0;
            
            num_clock_thread0 = 32'd0;
            num_clock_thread1 = 32'd0;
            num_clock_thread2 = 32'd0;
            num_clock_thread3 = 32'd0;  
            count_per_instr = 6'b0;           
            
            latency_thread0 = 32'd0;
            latency_thread1 = 32'd0;
            latency_thread2 = 32'd0;
            latency_thread3 = 32'd0;
            
            latency_total = 32'd0;    
`endif	
			
			thread = 0;
			pc = 0;
			sp = 0;
			sr = 0;
			
			for (i = 0; i < THREAD_COUNT; i = i + 1) begin
				state[i] = `INACTIVE;
				pcs[i] = 0;
				srs[i] = 0;
				sps[i] = 0;
			end

			run_vo = 1'b0;

			// Reset triggers interrupt in first thread
			state[0] = `INTERRUPT;
			int_vo = 1'b1;

			pc = 16'hFFFE;
			vector = 4'd15;
		end
		else begin
			// Update from completed instructions first
			`ifdef SIM
			    num_clock = num_clock + 1;
			    count_per_instr = count_per_instr + 1;
			    
			    
    
            `endif	
			
			if (done) begin
			    count_per_instr = 0;
				pcs[threadn] = pcn;
				sps[threadn] = spn;
				srs[threadn] = srn;
				// Leave inactive threads inactive
				if (state[threadn] != `INACTIVE)
					state[threadn] = `IDLE;			

`ifdef SIM
				icount = icount + 1;
				
				if(threadn == 0)
				    icount_thread0 = icount_thread0 + 1;
				else if(threadn == 1)
				    icount_thread1 = icount_thread1 + 1;
                else if(threadn == 2)
				    icount_thread2 = icount_thread2 + 1;
                else if(threadn == 3)
				    icount_thread3 = icount_thread3 + 1;	
				    
				    
				if(threadn == 0) begin
				    latency_total = latency_thread0 + latency_total - 1;
				    latency_thread0 = 0;
				    num_clock_thread0 = num_clock_thread0 - 1;
				end
				else if(threadn == 1) begin
				    latency_total = latency_thread1 + latency_total - 1;
				    latency_thread1 = 0;
				    num_clock_thread1 = num_clock_thread1 - 1;
				end
                else if(threadn == 2) begin
				    latency_total = latency_thread2 + latency_total - 1;
				    latency_thread2 = 0;
				    num_clock_thread2 = num_clock_thread2 - 1;
				end
                else if(threadn == 3) begin
				    latency_total = latency_thread3 + latency_total - 1;    			    
    				latency_thread3 = 0;
    				num_clock_thread3 = num_clock_thread3 - 1;	
				end
`endif	
			end
			
`ifdef SIM
			if(state[0] == `RUNNING) begin
				latency_thread0 = latency_thread0 +1;
				num_clock_thread0 = num_clock_thread0 + 1;
			end
			if(state[1] == `RUNNING) begin
				 latency_thread1 = latency_thread1 +1;
				 num_clock_thread1 = num_clock_thread1 + 1;
			end
            if(state[2] == `RUNNING) begin
				 latency_thread2 = latency_thread2 +1;
				 num_clock_thread2 = num_clock_thread2 + 1;
			end
            if(state[3] == `RUNNING) begin
				 latency_thread3 = latency_thread3 +1;
				 num_clock_thread3 = num_clock_thread3 + 1;	
			end
`endif	

			int_vo = 1'b0;

			if (state[thread] == `INTERRUPT) begin
				int_vo = 1'b1;
				state[thread] = `RUNNING;
			end 
			else begin
				// Launch thread if target is idle
				if (!run_si) begin
					run_vo = 1'b0;

					next_thread = thread + 1;
					if (state[next_thread] != `IDLE) begin
						next_thread = next_thread + 1;
						if (state[next_thread] != `IDLE) begin
							next_thread = next_thread + 1;
						end
					end
					
					if (state[next_thread] == `IDLE)
						thread = next_thread;

					if (state[thread] == `IDLE) begin
						pc = pcs[thread];
						sp = sps[thread];
						sr = srs[thread];
						
						run_vo = 1'b1;
						
						state[thread] = `RUNNING;
					end
				end
			end

			// Handle memory mapped interface
			if (en && we) begin
				if (address == `THREAD0_PC) begin
	`ifdef SIM
					$display("setting pc0 = %h", datain);
	`endif
					pcs[0] = datain;
				end
				else if (address == `THREAD1_PC) begin
	`ifdef SIM
					$display("setting pc1 = %h", datain);
	`endif
					pcs[1] = datain;
				end
				else if (address == `THREAD2_PC) begin
	`ifdef SIM
					$display("setting pc2 = %h", datain);
	`endif
					pcs[2] = datain;
				end
				else if (address == `THREAD3_PC) begin
	`ifdef SIM
					$display("setting pc3 = %h", datain);
	`endif
					pcs[3] = datain;
				end
				else if (address == `THREAD0_SP) begin
	`ifdef SIM
					$display("setting sp0 = %h", datain);
	`endif
					sps[0] = datain;
				end
				else if (address == `THREAD1_SP) begin
	`ifdef SIM
					$display("setting sp1 = %h", datain);
	`endif
					sps[1] = datain;
				end
				else if (address == `THREAD2_SP) begin
	`ifdef SIM
					$display("setting sp2 = %h", datain);
	`endif
					sps[2] = datain;
				end
				else if (address == `THREAD3_SP) begin
	`ifdef SIM
					$display("setting sp3 = %h", datain);
	`endif
					sps[3] = datain;
				end
				else if (address == `THREAD0_ACTIVE) begin
					if (datain == 16'd0) begin
						state[0] = `INACTIVE;
					end
					else if (state[0] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread0");
	`endif
						state[0] = `IDLE;
					end
				end
				else if (address == `THREAD1_ACTIVE) begin
					if (datain == 16'd0) begin
						state[1] = `INACTIVE;
					end
					else if (state[1] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread1");
	`endif
						state[1] = `IDLE;
					end
				end
				else if (address == `THREAD2_ACTIVE) begin
					if (datain == 16'd0) begin
						state[2] = `INACTIVE;
					end
					else if (state[2] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread2");
	`endif
						state[2] = `IDLE;
					end
				end
				else if (address == `THREAD3_ACTIVE) begin
					if (datain == 16'd0) begin
						state[3] = `INACTIVE;
					end
					else if (state[3] == `INACTIVE) begin
	`ifdef SIM
						$display("activating thread3");
	`endif
						state[3] = `IDLE;
					end
				end
			end
		end
	end
	//
	//////////////////////////////////////////////////////////////////////////////////////////////
	
endmodule
